@@ -17,15 +17,15 @@ import (
 	"context"
 	"errors"
 	"fmt"
+	"log/slog"
 	"net/url"
 	"slices"
 	"strings"
 	"sync"
 	"time"
 
-	"github.com/go-kit/log"
-	"github.com/go-kit/log/level"
 	"github.com/prometheus/client_golang/prometheus"
+	"github.com/prometheus/common/promslog"
 	"golang.org/x/sync/semaphore"
 
 	"github.com/prometheus/prometheus/model/labels"
@@ -96,7 +96,7 @@ type Manager struct {
 	done     chan struct{}
 	restored bool
 
-	logger log.Logger
+	logger *slog.Logger
 }
 
 // NotifyFunc sends notifications about a set of alerts generated by the given expression.
@@ -110,7 +110,7 @@ type ManagerOptions struct {
 	Context                   context.Context
 	Appendable                storage.Appendable
 	Queryable                 storage.Queryable
-	Logger                    log.Logger
+	Logger                    *slog.Logger
 	Registerer                prometheus.Registerer
 	OutageTolerance           time.Duration
 	ForGracePeriod            time.Duration
@@ -148,6 +148,10 @@ func NewManager(o *ManagerOptions) *Manager {
 		o.RuleDependencyController = ruleDependencyController{}
 	}
 
+	if o.Logger == nil {
+		o.Logger = promslog.NewNopLogger()
+	}
+
 	m := &Manager{
 		groups: map[string]*Group{},
 		opts:   o,
@@ -161,7 +165,7 @@ func NewManager(o *ManagerOptions) *Manager {
 
 // Run starts processing of the rule manager. It is blocking.
 func (m *Manager) Run() {
-	level.Info(m.logger).Log("msg", "Starting rule manager...")
+	m.logger.Info("Starting rule manager...")
 	m.start()
 	<-m.done
 }
@@ -175,7 +179,7 @@ func (m *Manager) Stop() {
 	m.mtx.Lock()
 	defer m.mtx.Unlock()
 
-	level.Info(m.logger).Log("msg", "Stopping rule manager...")
+	m.logger.Info("Stopping rule manager...")
 
 	for _, eg := range m.groups {
 		eg.stop()
@@ -185,7 +189,7 @@ func (m *Manager) Stop() {
 	// staleness markers.
 	close(m.done)
 
-	level.Info(m.logger).Log("msg", "Rule manager stopped")
+	m.logger.Info("Rule manager stopped")
 }
 
 // Update the rule manager's state as the config requires. If
@@ -206,7 +210,7 @@ func (m *Manager) Update(interval time.Duration, files []string, externalLabels
 
 	if errs != nil {
 		for _, e := range errs {
-			level.Error(m.logger).Log("msg", "loading groups failed", "err", e)
+			m.logger.Error("loading groups failed", "err", e)
 		}
 		return errors.New("error loading rules, previous rule set restored")
 	}
@@ -312,25 +316,27 @@ func (m *Manager) LoadGroups(
 					return nil, []error{fmt.Errorf("%s: %w", fn, err)}
 				}
 
+				mLabels := FromMaps(rg.Labels, r.Labels)
+
 				if r.Alert.Value != "" {
 					rules = append(rules, NewAlertingRule(
 						r.Alert.Value,
 						expr,
 						time.Duration(r.For),
 						time.Duration(r.KeepFiringFor),
-						labels.FromMap(r.Labels),
+						mLabels,
 						labels.FromMap(r.Annotations),
 						externalLabels,
 						externalURL,
 						m.restored,
-						log.With(m.logger, "alert", r.Alert),
+						m.logger.With("alert", r.Alert),
 					))
 					continue
 				}
 				rules = append(rules, NewRecordingRule(
 					r.Record.Value,
 					expr,
-					labels.FromMap(r.Labels),
+					mLabels,
 				))
 			}
 
@@ -501,3 +507,16 @@ func (c sequentialRuleEvalController) Allow(_ context.Context, _ *Group, _ Rule)
 }
 
 func (c sequentialRuleEvalController) Done(_ context.Context) {}
+
+// FromMaps returns new sorted Labels from the given maps, overriding each other in order.
+func FromMaps(maps ...map[string]string) labels.Labels {
+	mLables := make(map[string]string)
+
+	for _, m := range maps {
+		for k, v := range m {
+			mLables[k] = v
+		}
+	}
+
+	return labels.FromMap(mLables)
+}
