@@ -14,6 +14,7 @@
 package scrape
 
 import (
+	"bytes"
 	"context"
 	"fmt"
 	"net/http"
@@ -26,33 +27,42 @@ import (
 	"testing"
 	"time"
 
-	"github.com/go-kit/log"
 	"github.com/gogo/protobuf/proto"
 	"github.com/prometheus/client_golang/prometheus"
 	dto "github.com/prometheus/client_model/go"
+	"github.com/prometheus/common/expfmt"
 	"github.com/prometheus/common/model"
+	"github.com/prometheus/common/promslog"
 	"github.com/stretchr/testify/require"
 	"google.golang.org/protobuf/types/known/timestamppb"
 	"gopkg.in/yaml.v2"
 
+	"github.com/prometheus/prometheus/model/timestamp"
+
 	"github.com/prometheus/prometheus/config"
 	"github.com/prometheus/prometheus/discovery"
 	_ "github.com/prometheus/prometheus/discovery/file"
 	"github.com/prometheus/prometheus/discovery/targetgroup"
+	"github.com/prometheus/prometheus/model/histogram"
 	"github.com/prometheus/prometheus/model/labels"
 	"github.com/prometheus/prometheus/model/relabel"
+	"github.com/prometheus/prometheus/tsdb/tsdbutil"
 	"github.com/prometheus/prometheus/util/runutil"
 	"github.com/prometheus/prometheus/util/testutil"
 )
 
+func init() {
+	// This can be removed when the default validation scheme in common is updated.
+	model.NameValidationScheme = model.UTF8Validation
+}
+
 func TestPopulateLabels(t *testing.T) {
 	cases := []struct {
-		in            labels.Labels
-		cfg           *config.ScrapeConfig
-		noDefaultPort bool
-		res           labels.Labels
-		resOrig       labels.Labels
-		err           string
+		in      labels.Labels
+		cfg     *config.ScrapeConfig
+		res     labels.Labels
+		resOrig labels.Labels
+		err     string
 	}{
 		// Regular population of scrape config options.
 		{
@@ -106,8 +116,8 @@ func TestPopulateLabels(t *testing.T) {
 				ScrapeTimeout:  model.Duration(time.Second),
 			},
 			res: labels.FromMap(map[string]string{
-				model.AddressLabel:        "1.2.3.4:80",
-				model.InstanceLabel:       "1.2.3.4:80",
+				model.AddressLabel:        "1.2.3.4",
+				model.InstanceLabel:       "1.2.3.4",
 				model.SchemeLabel:         "http",
 				model.MetricsPathLabel:    "/custom",
 				model.JobLabel:            "custom-job",
@@ -137,7 +147,7 @@ func TestPopulateLabels(t *testing.T) {
 				ScrapeTimeout:  model.Duration(time.Second),
 			},
 			res: labels.FromMap(map[string]string{
-				model.AddressLabel:        "[::1]:443",
+				model.AddressLabel:        "[::1]",
 				model.InstanceLabel:       "custom-instance",
 				model.SchemeLabel:         "https",
 				model.MetricsPathLabel:    "/metrics",
@@ -360,7 +370,6 @@ func TestPopulateLabels(t *testing.T) {
 				ScrapeInterval: model.Duration(time.Second),
 				ScrapeTimeout:  model.Duration(time.Second),
 			},
-			noDefaultPort: true,
 			res: labels.FromMap(map[string]string{
 				model.AddressLabel:        "1.2.3.4",
 				model.InstanceLabel:       "1.2.3.4",
@@ -379,7 +388,7 @@ func TestPopulateLabels(t *testing.T) {
 				model.ScrapeTimeoutLabel:  "1s",
 			}),
 		},
-		// Remove default port (http).
+		// verify that the default port is not removed (http).
 		{
 			in: labels.FromMap(map[string]string{
 				model.AddressLabel: "1.2.3.4:80",
@@ -391,9 +400,8 @@ func TestPopulateLabels(t *testing.T) {
 				ScrapeInterval: model.Duration(time.Second),
 				ScrapeTimeout:  model.Duration(time.Second),
 			},
-			noDefaultPort: true,
 			res: labels.FromMap(map[string]string{
-				model.AddressLabel:        "1.2.3.4",
+				model.AddressLabel:        "1.2.3.4:80",
 				model.InstanceLabel:       "1.2.3.4:80",
 				model.SchemeLabel:         "http",
 				model.MetricsPathLabel:    "/metrics",
@@ -410,7 +418,7 @@ func TestPopulateLabels(t *testing.T) {
 				model.ScrapeTimeoutLabel:  "1s",
 			}),
 		},
-		// Remove default port (https).
+		// verify that the default port is not removed (https).
 		{
 			in: labels.FromMap(map[string]string{
 				model.AddressLabel: "1.2.3.4:443",
@@ -422,9 +430,8 @@ func TestPopulateLabels(t *testing.T) {
 				ScrapeInterval: model.Duration(time.Second),
 				ScrapeTimeout:  model.Duration(time.Second),
 			},
-			noDefaultPort: true,
 			res: labels.FromMap(map[string]string{
-				model.AddressLabel:        "1.2.3.4",
+				model.AddressLabel:        "1.2.3.4:443",
 				model.InstanceLabel:       "1.2.3.4:443",
 				model.SchemeLabel:         "https",
 				model.MetricsPathLabel:    "/metrics",
@@ -445,7 +452,7 @@ func TestPopulateLabels(t *testing.T) {
 	for _, c := range cases {
 		in := c.in.Copy()
 
-		res, orig, err := PopulateLabels(labels.NewBuilder(c.in), c.cfg, c.noDefaultPort)
+		res, orig, err := PopulateLabels(labels.NewBuilder(c.in), c.cfg)
 		if c.err != "" {
 			require.EqualError(t, err, c.err)
 		} else {
@@ -716,37 +723,256 @@ scrape_configs:
 	require.ElementsMatch(t, []string{"job1", "job3"}, scrapeManager.ScrapePools())
 }
 
-// TestManagerCTZeroIngestion tests scrape manager for CT cases.
+func setupScrapeManager(t *testing.T, honorTimestamps, enableCTZeroIngestion bool) (*collectResultAppender, *Manager) {
+	app := &collectResultAppender{}
+	scrapeManager, err := NewManager(
+		&Options{
+			EnableCreatedTimestampZeroIngestion: enableCTZeroIngestion,
+			skipOffsetting:                      true,
+		},
+		promslog.New(&promslog.Config{}),
+		nil,
+		&collectResultAppendable{app},
+		prometheus.NewRegistry(),
+	)
+	require.NoError(t, err)
+
+	require.NoError(t, scrapeManager.ApplyConfig(&config.Config{
+		GlobalConfig: config.GlobalConfig{
+			// Disable regular scrapes.
+			ScrapeInterval:  model.Duration(9999 * time.Minute),
+			ScrapeTimeout:   model.Duration(5 * time.Second),
+			ScrapeProtocols: []config.ScrapeProtocol{config.OpenMetricsText1_0_0, config.PrometheusProto},
+		},
+		ScrapeConfigs: []*config.ScrapeConfig{{JobName: "test", HonorTimestamps: honorTimestamps}},
+	}))
+
+	return app, scrapeManager
+}
+
+func setupTestServer(t *testing.T, typ string, toWrite []byte) *httptest.Server {
+	once := sync.Once{}
+
+	server := httptest.NewServer(
+		http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
+			fail := true
+			once.Do(func() {
+				fail = false
+				w.Header().Set("Content-Type", typ)
+				w.Write(toWrite)
+			})
+
+			if fail {
+				w.WriteHeader(http.StatusInternalServerError)
+			}
+		}),
+	)
+
+	t.Cleanup(func() { server.Close() })
+
+	return server
+}
+
+// TestManagerCTZeroIngestion tests scrape manager for various CT cases.
 func TestManagerCTZeroIngestion(t *testing.T) {
-	const mName = "expected_counter"
+	const (
+		// _total suffix is required, otherwise expfmt with OMText will mark metric as "unknown"
+		expectedMetricName        = "expected_metric_total"
+		expectedCreatedMetricName = "expected_metric_created"
+		expectedSampleValue       = 17.0
+	)
+
+	for _, testFormat := range []config.ScrapeProtocol{config.PrometheusProto, config.OpenMetricsText1_0_0} {
+		t.Run(fmt.Sprintf("format=%s", testFormat), func(t *testing.T) {
+			for _, testWithCT := range []bool{false, true} {
+				t.Run(fmt.Sprintf("withCT=%v", testWithCT), func(t *testing.T) {
+					for _, testCTZeroIngest := range []bool{false, true} {
+						t.Run(fmt.Sprintf("ctZeroIngest=%v", testCTZeroIngest), func(t *testing.T) {
+							sampleTs := time.Now()
+							ctTs := time.Time{}
+							if testWithCT {
+								ctTs = sampleTs.Add(-2 * time.Minute)
+							}
+
+							// TODO(bwplotka): Add more types than just counter?
+							encoded := prepareTestEncodedCounter(t, testFormat, expectedMetricName, expectedSampleValue, sampleTs, ctTs)
+							app, scrapeManager := setupScrapeManager(t, true, testCTZeroIngest)
+
+							// Perform the test.
+							doOneScrape(t, scrapeManager, app, setupTestServer(t, config.ScrapeProtocolsHeaders[testFormat], encoded))
+
+							// Verify results.
+							// Verify what we got vs expectations around CT injection.
+							samples := findSamplesForMetric(app.resultFloats, expectedMetricName)
+							if testWithCT && testCTZeroIngest {
+								require.Len(t, samples, 2)
+								require.Equal(t, 0.0, samples[0].f)
+								require.Equal(t, timestamp.FromTime(ctTs), samples[0].t)
+								require.Equal(t, expectedSampleValue, samples[1].f)
+								require.Equal(t, timestamp.FromTime(sampleTs), samples[1].t)
+							} else {
+								require.Len(t, samples, 1)
+								require.Equal(t, expectedSampleValue, samples[0].f)
+								require.Equal(t, timestamp.FromTime(sampleTs), samples[0].t)
+							}
+
+							// Verify what we got vs expectations around additional _created series for OM text.
+							// enableCTZeroInjection also kills that _created line.
+							createdSeriesSamples := findSamplesForMetric(app.resultFloats, expectedCreatedMetricName)
+							if testFormat == config.OpenMetricsText1_0_0 && testWithCT && !testCTZeroIngest {
+								// For OM Text, when counter has CT, and feature flag disabled we should see _created lines.
+								require.Len(t, createdSeriesSamples, 1)
+								// Conversion taken from common/expfmt.writeOpenMetricsFloat.
+								// We don't check the ct timestamp as explicit ts was not implemented in expfmt.Encoder,
+								// but exists in OM https://github.com/OpenObservability/OpenMetrics/blob/main/specification/OpenMetrics.md#:~:text=An%20example%20with%20a%20Metric%20with%20no%20labels%2C%20and%20a%20MetricPoint%20with%20a%20timestamp%20and%20a%20created
+								// We can implement this, but we want to potentially get rid of OM 1.0 CT lines
+								require.Equal(t, float64(timestamppb.New(ctTs).AsTime().UnixNano())/1e9, createdSeriesSamples[0].f)
+							} else {
+								require.Empty(t, createdSeriesSamples)
+							}
+						})
+					}
+				})
+			}
+		})
+	}
+}
+
+func prepareTestEncodedCounter(t *testing.T, format config.ScrapeProtocol, mName string, v float64, ts, ct time.Time) (encoded []byte) {
+	t.Helper()
+
+	counter := &dto.Counter{Value: proto.Float64(v)}
+	if !ct.IsZero() {
+		counter.CreatedTimestamp = timestamppb.New(ct)
+	}
+	ctrType := dto.MetricType_COUNTER
+	inputMetric := &dto.MetricFamily{
+		Name: proto.String(mName),
+		Type: &ctrType,
+		Metric: []*dto.Metric{{
+			TimestampMs: proto.Int64(timestamp.FromTime(ts)),
+			Counter:     counter,
+		}},
+	}
+	switch format {
+	case config.PrometheusProto:
+		return protoMarshalDelimited(t, inputMetric)
+	case config.OpenMetricsText1_0_0:
+		buf := &bytes.Buffer{}
+		require.NoError(t, expfmt.NewEncoder(buf, expfmt.NewFormat(expfmt.TypeOpenMetrics), expfmt.WithCreatedLines(), expfmt.WithUnit()).Encode(inputMetric))
+		_, _ = buf.WriteString("# EOF")
+
+		t.Log("produced OM text to expose:", buf.String())
+		return buf.Bytes()
+	default:
+		t.Fatalf("not implemented format: %v", format)
+		return nil
+	}
+}
+
+func doOneScrape(t *testing.T, manager *Manager, appender *collectResultAppender, server *httptest.Server) {
+	t.Helper()
+
+	serverURL, err := url.Parse(server.URL)
+	require.NoError(t, err)
+
+	// Add fake target directly into tsets + reload
+	manager.updateTsets(map[string][]*targetgroup.Group{
+		"test": {{
+			Targets: []model.LabelSet{{
+				model.SchemeLabel:  model.LabelValue(serverURL.Scheme),
+				model.AddressLabel: model.LabelValue(serverURL.Host),
+			}},
+		}},
+	})
+	manager.reload()
+
+	// Wait for one scrape.
+	ctx, cancel := context.WithTimeout(context.Background(), 1*time.Minute)
+	defer cancel()
+	require.NoError(t, runutil.Retry(100*time.Millisecond, ctx.Done(), func() error {
+		appender.mtx.Lock()
+		defer appender.mtx.Unlock()
+
+		// Check if scrape happened and grab the relevant samples.
+		if len(appender.resultFloats) > 0 {
+			return nil
+		}
+		return fmt.Errorf("expected some float samples, got none")
+	}), "after 1 minute")
+	manager.Stop()
+}
+
+func findSamplesForMetric(floats []floatSample, metricName string) (ret []floatSample) {
+	for _, f := range floats {
+		if f.metric.Get(model.MetricNameLabel) == metricName {
+			ret = append(ret, f)
+		}
+	}
+	return ret
+}
+
+// generateTestHistogram generates the same thing as tsdbutil.GenerateTestHistogram,
+// but in the form of dto.Histogram.
+func generateTestHistogram(i int) *dto.Histogram {
+	helper := tsdbutil.GenerateTestHistogram(i)
+	h := &dto.Histogram{}
+	h.SampleCount = proto.Uint64(helper.Count)
+	h.SampleSum = proto.Float64(helper.Sum)
+	h.Schema = proto.Int32(helper.Schema)
+	h.ZeroThreshold = proto.Float64(helper.ZeroThreshold)
+	h.ZeroCount = proto.Uint64(helper.ZeroCount)
+	h.PositiveSpan = make([]*dto.BucketSpan, len(helper.PositiveSpans))
+	for i, span := range helper.PositiveSpans {
+		h.PositiveSpan[i] = &dto.BucketSpan{
+			Offset: proto.Int32(span.Offset),
+			Length: proto.Uint32(span.Length),
+		}
+	}
+	h.PositiveDelta = helper.PositiveBuckets
+	h.NegativeSpan = make([]*dto.BucketSpan, len(helper.NegativeSpans))
+	for i, span := range helper.NegativeSpans {
+		h.NegativeSpan[i] = &dto.BucketSpan{
+			Offset: proto.Int32(span.Offset),
+			Length: proto.Uint32(span.Length),
+		}
+	}
+	h.NegativeDelta = helper.NegativeBuckets
+	return h
+}
+
+func TestManagerCTZeroIngestionHistogram(t *testing.T) {
+	const mName = "expected_histogram"
 
 	for _, tc := range []struct {
 		name                  string
-		counterSample         *dto.Counter
+		inputHistSample       *dto.Histogram
 		enableCTZeroIngestion bool
 	}{
 		{
-			name: "disabled with CT on counter",
-			counterSample: &dto.Counter{
-				Value: proto.Float64(1.0),
-				// Timestamp does not matter as long as it exists in this test.
-				CreatedTimestamp: timestamppb.Now(),
-			},
+			name: "disabled with CT on histogram",
+			inputHistSample: func() *dto.Histogram {
+				h := generateTestHistogram(0)
+				h.CreatedTimestamp = timestamppb.Now()
+				return h
+			}(),
+			enableCTZeroIngestion: false,
 		},
 		{
-			name: "enabled with CT on counter",
-			counterSample: &dto.Counter{
-				Value: proto.Float64(1.0),
-				// Timestamp does not matter as long as it exists in this test.
-				CreatedTimestamp: timestamppb.Now(),
-			},
+			name: "enabled with CT on histogram",
+			inputHistSample: func() *dto.Histogram {
+				h := generateTestHistogram(0)
+				h.CreatedTimestamp = timestamppb.Now()
+				return h
+			}(),
 			enableCTZeroIngestion: true,
 		},
 		{
-			name: "enabled without CT on counter",
-			counterSample: &dto.Counter{
-				Value: proto.Float64(1.0),
-			},
+			name: "enabled without CT on histogram",
+			inputHistSample: func() *dto.Histogram {
+				h := generateTestHistogram(0)
+				return h
+			}(),
 			enableCTZeroIngestion: true,
 		},
 	} {
@@ -755,9 +981,10 @@ func TestManagerCTZeroIngestion(t *testing.T) {
 			scrapeManager, err := NewManager(
 				&Options{
 					EnableCreatedTimestampZeroIngestion: tc.enableCTZeroIngestion,
+					EnableNativeHistogramsIngestion:     true,
 					skipOffsetting:                      true,
 				},
-				log.NewLogfmtLogger(os.Stderr),
+				promslog.New(&promslog.Config{}),
 				nil,
 				&collectResultAppendable{app},
 				prometheus.NewRegistry(),
@@ -780,16 +1007,16 @@ func TestManagerCTZeroIngestion(t *testing.T) {
 			// Start fake HTTP target to that allow one scrape only.
 			server := httptest.NewServer(
 				http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
-					fail := true
+					fail := true // TODO(bwplotka): Kill or use?
 					once.Do(func() {
 						fail = false
 						w.Header().Set("Content-Type", `application/vnd.google.protobuf; proto=io.prometheus.client.MetricFamily; encoding=delimited`)
 
-						ctrType := dto.MetricType_COUNTER
+						ctrType := dto.MetricType_HISTOGRAM
 						w.Write(protoMarshalDelimited(t, &dto.MetricFamily{
 							Name:   proto.String(mName),
 							Type:   &ctrType,
-							Metric: []*dto.Metric{{Counter: tc.counterSample}},
+							Metric: []*dto.Metric{{Histogram: tc.inputHistSample}},
 						}))
 					})
 
@@ -815,7 +1042,8 @@ func TestManagerCTZeroIngestion(t *testing.T) {
 			})
 			scrapeManager.reload()
 
-			var got []float64
+			var got []histogramSample
+
 			// Wait for one scrape.
 			ctx, cancel := context.WithTimeout(context.Background(), 1*time.Minute)
 			defer cancel()
@@ -823,32 +1051,35 @@ func TestManagerCTZeroIngestion(t *testing.T) {
 				app.mtx.Lock()
 				defer app.mtx.Unlock()
 
-				// Check if scrape happened and grab the relevant samples, they have to be there - or it's a bug
+				// Check if scrape happened and grab the relevant histograms, they have to be there - or it's a bug
 				// and it's not worth waiting.
-				for _, f := range app.resultFloats {
-					if f.metric.Get(model.MetricNameLabel) == mName {
-						got = append(got, f.f)
+				for _, h := range app.resultHistograms {
+					if h.metric.Get(model.MetricNameLabel) == mName {
+						got = append(got, h)
 					}
 				}
-				if len(app.resultFloats) > 0 {
+				if len(app.resultHistograms) > 0 {
 					return nil
 				}
-				return fmt.Errorf("expected some samples, got none")
+				return fmt.Errorf("expected some histogram samples, got none")
 			}), "after 1 minute")
 			scrapeManager.Stop()
 
-			// Check for zero samples, assuming we only injected always one sample.
+			// Check for zero samples, assuming we only injected always one histogram sample.
 			// Did it contain CT to inject? If yes, was CT zero enabled?
-			if tc.counterSample.CreatedTimestamp.IsValid() && tc.enableCTZeroIngestion {
+			if tc.inputHistSample.CreatedTimestamp.IsValid() && tc.enableCTZeroIngestion {
 				require.Len(t, got, 2)
-				require.Equal(t, 0.0, got[0])
-				require.Equal(t, tc.counterSample.GetValue(), got[1])
+				// Zero sample.
+				require.Equal(t, histogram.Histogram{}, *got[0].h)
+				// Quick soft check to make sure it's the same sample or at least not zero.
+				require.Equal(t, tc.inputHistSample.GetSampleSum(), got[1].h.Sum)
 				return
 			}
 
 			// Expect only one, valid sample.
 			require.Len(t, got, 1)
-			require.Equal(t, tc.counterSample.GetValue(), got[0])
+			// Quick soft check to make sure it's the same sample or at least not zero.
+			require.Equal(t, tc.inputHistSample.GetSampleSum(), got[0].h.Sum)
 		})
 	}
 }
@@ -894,7 +1125,7 @@ func runManagers(t *testing.T, ctx context.Context) (*discovery.Manager, *Manage
 	require.NoError(t, err)
 	discoveryManager := discovery.NewManager(
 		ctx,
-		log.NewNopLogger(),
+		promslog.NewNopLogger(),
 		reg,
 		sdMetrics,
 		discovery.Updatert(100*time.Millisecond),
@@ -1181,7 +1412,7 @@ scrape_configs:
 }
 
 // TestOnlyStaleTargetsAreDropped makes sure that when a job has multiple providers, when one of them should no
-// longer discover targets, only the stale targets of that provier are dropped.
+// longer discover targets, only the stale targets of that provider are dropped.
 func TestOnlyStaleTargetsAreDropped(t *testing.T) {
 	ctx, cancel := context.WithCancel(context.Background())
 	defer cancel()
